
---------------------------------------------------
-- This Design is optimized for Trenz TE0304carrier
-- with a TE630 Spartan LX 75 module mounted
-- 100 MHz OSC - clock for this module is 50 MHz
---------------------------------------------------


library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_unsigned.all;
  use ieee.numeric_std.all;
  
--  library UNISIM;
--  use UNISIM.Vcomponents.all;

entity VIDEO_OUT is
port (
    RESET       	: in    std_logic;
	 PIXEL_CLOCK	: in    std_logic;
    --
    VIDEO_R			: out   std_logic_vector (4 downto 0);
    VIDEO_G    	: out   std_logic_vector (4 downto 0);
    VIDEO_B    	: out   std_logic_vector (4 downto 0);
    HSYNC       	: out   std_logic;
    VSYNC       	: out   std_logic
    );
end;


architecture a of VIDEO_OUT is

   -- values for 800 x 600 @ 50 MHz
   
   constant V_FRONT_PORCH_START 	: std_logic_vector (11 downto 0) := x"258"; 		-- line 600
   constant V_SYNC_START    		: std_logic_vector (11 downto 0) := x"27D"; 		-- line 637
   constant V_BACK_PORCH_START  	: std_logic_vector (11 downto 0) := x"283";		-- line 643
   constant LINE_PER_FRAME      	: std_logic_vector (11 downto 0) := x"29A";		-- 666 lines

   constant H_FRONT_PORCH_START	: std_logic_vector (11 downto 0) := x"320";		-- pixel 800
   constant H_SYNC_START			: std_logic_vector (11 downto 0) := x"358";		-- pixel 856
   constant H_BACK_PORCH_START	: std_logic_vector (11 downto 0) := x"3D0";		-- pixel 976
   constant PIXEL_PER_LINE			: std_logic_vector (11 downto 0) := x"410";		-- 1040 pixels
  

   --Signals
   signal pixel_count		: std_logic_vector(10 downto 0);
   signal line_count			: std_logic_vector( 9 downto 0);

   signal pixel_value		: std_logic_vector(10 downto 0);
   signal line_value			: std_logic_vector( 9 downto 0);

   signal tempmul				: std_logic_vector(20 downto 0);
   signal image_value		: std_logic_vector( 4 downto 0);		-- 31 cols!
   signal color_value		: std_logic_vector( 4 downto 0);		-- final colour
   signal color_code			: std_logic_vector( 5 downto 0);		-- hw colour to toggle
  
   signal hor_end				: boolean;
   signal ver_end				: boolean;
  
   signal v_sync				: std_logic;
   signal h_sync				: std_logic;
   signal v_sync_d1			: std_logic;
   signal h_sync_d1			: std_logic;
   signal v_sync_d2			: std_logic;
   signal h_sync_d2			: std_logic;
   signal v_sync_d3			: std_logic;
   signal h_sync_d3			: std_logic;
   signal v_sync_d4			: std_logic;
   signal h_sync_d4			: std_logic;
   signal v_sync_d5			: std_logic;
   signal h_sync_d5			: std_logic;
  
   signal v_blank				: std_logic;
   signal h_blank  			: std_logic;
  
   signal gitter				: std_logic;
   signal frame				: std_logic;

   signal image_count		: std_logic_vector(33 downto 0);
   signal ball_count			: std_logic_vector(17 downto 0);
 
   signal ball 				: std_logic; 
   signal ball_x_pos			: std_logic_vector(10 downto 0);
   signal ball_y_pos			: std_logic_vector( 9 downto 0);
   signal ball_x_dir			: std_logic;
   signal ball_y_dir			: std_logic;
   signal ball_h_x_dif		: std_logic_vector(10 downto 0);
   signal ball_v_y_dif		: std_logic_vector( 9 downto 0);



begin


hor_end <= (pixel_count = (PIXEL_PER_LINE(10 downto 0) - "1"));
ver_end <= ( line_count = (LINE_PER_FRAME( 9 downto 0) - "1"));


p_h_cnt : process(PIXEL_CLOCK)
begin

	if rising_edge(PIXEL_CLOCK) then
		if (RESET = '1') then
			pixel_count <= (others => '0');
		
		else
			if hor_end then
			  pixel_count <= (others => '0');
			else
			  pixel_count <= pixel_count + "1";
			end if;
		end if;
		
	end if;	 

end process;


p_v_cnt: process(PIXEL_CLOCK)
begin
	if rising_edge(PIXEL_CLOCK) then
	
		if (RESET = '1') then
			line_count <= (others => '0');
			
		else
			if hor_end then
				if ver_end then
					line_count <= (others => '0');
				else
					line_count <= line_count + "1";
				end if;
			end if;
			
		end if;
		
   end if;
end process;

  
mk_v_sync : process(PIXEL_CLOCK)
	variable vcnt_is_front_porch_start : boolean;
	variable vcnt_is_sync_start        : boolean;
	variable vcnt_is_back_porch_start  : boolean;


begin


	if rising_edge(PIXEL_CLOCK) then
	

		if (RESET = '1') then
			v_sync <= '1';
			v_blank <= '0';
	
		else
	
			vcnt_is_front_porch_start		:= (line_count = (V_FRONT_PORCH_START(9 downto 0) - "1"));
			vcnt_is_sync_start       		:= (line_count = (       V_SYNC_START(9 downto 0) - "1"));
			vcnt_is_back_porch_start		:= (line_count = ( V_BACK_PORCH_START(9 downto 0) - "1"));

			if vcnt_is_sync_start and hor_end then
				v_sync <= '0';
			elsif vcnt_is_back_porch_start and hor_end then
				v_sync <= '1';
			end if;

			if vcnt_is_front_porch_start and hor_end then
				v_blank <= '1';
			elsif ver_end and hor_end then
				v_blank <= '0';
			end if;

		end if;
		
	end if;
end process;



mk_h_sync    : process(PIXEL_CLOCK)

	variable hcnt_is_front_porch_start     : boolean;
	variable hcnt_is_sync_start           	: boolean;
	variable hcnt_is_back_porch_start      : boolean;
	
begin
	if rising_edge(PIXEL_CLOCK) then
		
		if (RESET = '1') then
			h_sync <= '1';
			h_blank <= '0';
			
		else		
			hcnt_is_front_porch_start     := (pixel_count = ( H_FRONT_PORCH_START(10 downto 0) - "1"));
			hcnt_is_sync_start            := (pixel_count = (        H_SYNC_START(10 downto 0) - "1"));
			hcnt_is_back_porch_start      := (pixel_count = (  H_BACK_PORCH_START(10 downto 0) - "1"));

			if hcnt_is_sync_start then
				h_sync <= '0';
			elsif hcnt_is_back_porch_start then
				h_sync <= '1';
			end if;

			if hcnt_is_front_porch_start then
				h_blank <= '1';
			elsif hor_end then
				h_blank <= '0';
			end if;
		end if;
		
	end if;
end process;



-- for colour rotation
mk_image_cnt: process(PIXEL_CLOCK)
begin
	if rising_edge(PIXEL_CLOCK) then
		if (RESET = '1') then
			image_count <= (others => '0');
		else
			image_count <= std_logic_vector( unsigned(image_count) + 1);
		end if;
   end if;
end process;





--------------------------
-- generate image	data
--------------------------
line_value  <= std_logic_vector( unsigned( line_count( 9 downto 0)) - unsigned(pixel_count(10 downto 2)));
pixel_value <= std_logic_vector( unsigned(pixel_count(10 downto 0)) - unsigned( line_count( 9 downto 2)));
tempmul <= std_logic_vector(unsigned(line_value) * unsigned(pixel_value) + unsigned(image_count(33 downto 13)));
image_value <= tempmul(18 downto 14);		


-- generate outer frame
-- according to delay it is not totally synched to x,y
-- possibly use 2 FFs more to delay signal "frame"

mk_frame : process(pixel_count,line_count)
begin
			if (unsigned(pixel_count(10 downto 0)) =   0) then
			frame <= '1';
		elsif (unsigned(pixel_count(10 downto 0)) = 799) then
			frame <= '1';
		elsif (unsigned( line_count( 9 downto 0)) =   0) then
			frame <= '1';
		elsif (unsigned( line_count( 9 downto 0)) = 599) then
			frame <= '1';
		else
			frame <= '0';
		end if;

end process;


-- generate a mesh pattern
-- according to delay it is not totally synched to x,y
-- possibly use 2 FFs more to delay signal "gitter"

mk_gitter : process(line_count,pixel_count)
begin
			if (unsigned( line_count(5 downto 0)) =   0) then
			gitter <= '1';
		elsif (unsigned( line_count(5 downto 0)) =   63) then
			gitter <= '1';
		elsif (unsigned(pixel_count(5 downto 0)) =   0) then
			gitter <= '1';
		elsif (unsigned(pixel_count(5 downto 0)) =   63) then
			gitter <= '1';
		else
			gitter <= '0';

	end if;
end process;


-- generate a cursor / ball
mk_ball : process(PIXEL_CLOCK)
begin
   if (rising_edge(PIXEL_CLOCK)) then
	
		if (RESET = '1') then
			ball_count <= (others => '0');
			
			ball_x_pos <= (others => '0');
			ball_y_pos <= (others => '0');
			
			ball_x_dir <= '1';
			ball_y_dir <= '1';
			
		else
			ball_count <= std_logic_vector( unsigned(ball_count) + 1);

			-- eventually move ball
			if (unsigned(ball_count) = 0) then
				if (ball_x_dir = '1') then
					ball_x_pos <= std_logic_vector( unsigned(ball_x_pos) + 1);
				else
					ball_x_pos <= std_logic_vector( unsigned(ball_x_pos) - 1);
				end if;

				if (ball_y_dir = '1') then
					ball_y_pos <= std_logic_vector( unsigned(ball_y_pos) + 1);
				else
					ball_y_pos <= std_logic_vector( unsigned(ball_y_pos) - 1);
				end if;
			end if;

			-- decide new directions if wall reflects ball
			if (unsigned(ball_x_pos) > 790) then
				ball_x_dir <= '0';
			elsif (unsigned(ball_x_pos) <  10) then
				ball_x_dir <= '1';
			else
				-- keep value
			end if;
			
			if (unsigned(ball_y_pos) > 590) then
				ball_y_dir <= '0';
			elsif (unsigned(ball_y_pos) <  10) then
				ball_y_dir <= '1';
			else
				-- keep value			
			end if;

			-- distance between ball and video beam
			if (unsigned(ball_x_pos) > unsigned(pixel_count)) then
				ball_h_x_dif <= std_logic_vector( unsigned(ball_x_pos) - unsigned(pixel_count) );
			else
				ball_h_x_dif <= std_logic_vector( unsigned(pixel_count) - unsigned(ball_x_pos) );
			end if;
			
			if (unsigned(ball_y_pos) > unsigned(line_count)) then
				ball_v_y_dif <= std_logic_vector( unsigned(ball_y_pos) - unsigned( line_count) );
			else
				ball_v_y_dif <= std_logic_vector( unsigned( line_count) - unsigned(ball_y_pos) );
			end if;

			-- decide if ball has to be displayed
			if ((unsigned(ball_v_y_dif) < 10) and (unsigned(ball_h_x_dif) <10 )) then
				ball <= '1';
			else
				ball <= '0';
			end if;
			
		end if;

	end if;
end process;



----------------------------------------
-- summarize all objects with priority
----------------------------------------
mk_cols  : process (PIXEL_CLOCK)
begin
	if rising_edge(PIXEL_CLOCK) then
		
		if (h_blank = '1' or v_blank = '1') then
			color_value <= "00000";
			
		elsif (ball = '1') then
			color_value <= "00011";
			
		elsif (frame = '1') then
			color_value <= image_count(29 downto 25);
		
		elsif (gitter = '1') then
			color_value <= std_logic_vector(31 - unsigned(image_value));
			
		else
			color_value <= image_value;
		end if;
	end if;
	
end process;



---------------------------------------------
-- create colour information from colour index
----------------------------------------------

mk_colours  : process (PIXEL_CLOCK)
begin
	 if rising_edge(PIXEL_CLOCK) then
	
		   if (color_value = "00000") then color_code <= "000000";	-- black
		elsif (color_value = "00001") then color_code <= "000001";	-- black/red
		elsif (color_value = "00010") then color_code <= "001001";	-- red  /red
		elsif (color_value = "00011") then color_code <= "001011";	-- red  /yellow
		elsif (color_value = "00100") then color_code <= "011011";	-- yellow
		elsif (color_value = "00101") then color_code <= "011010";	-- yellow/green
		elsif (color_value = "00110") then color_code <= "010010";	-- green
		elsif (color_value = "00111") then color_code <= "010110";	-- green/cyan
		elsif (color_value = "01000") then color_code <= "110110";	-- cyan
		elsif (color_value = "01001") then color_code <= "110100";	-- cyan/blue
		elsif (color_value = "01010") then color_code <= "100100";	-- blue
		elsif (color_value = "01011") then color_code <= "100101";	-- blue/purple
		elsif (color_value = "01100") then color_code <= "101101";	-- purple
		elsif (color_value = "01101") then color_code <= "101001";	-- purple/red
		elsif (color_value = "01110") then color_code <= "000000";	-- black
		elsif (color_value = "01111") then color_code <= "000000";	-- 
		elsif (color_value = "10000") then color_code <= "000001";	-- 1
		elsif (color_value = "10001") then color_code <= "000011";	-- 3
		elsif (color_value = "10010") then color_code <= "000010";	-- 2
		elsif (color_value = "10011") then color_code <= "000110";	-- 6
		elsif (color_value = "10100") then color_code <= "000100";	-- 4
		elsif (color_value = "10101") then color_code <= "000101";	-- 5
		elsif (color_value = "10110") then color_code <= "000000";	-- 
		elsif (color_value = "10111") then color_code <= "111000";	-- 
		elsif (color_value = "11000") then color_code <= "111001";	-- 1
		elsif (color_value = "11001") then color_code <= "111011";	-- 3
		elsif (color_value = "11010") then color_code <= "111010";	-- 2
		elsif (color_value = "11011") then color_code <= "111110";	-- 6
		elsif (color_value = "11100") then color_code <= "111100";	-- 4
		elsif (color_value = "11101") then color_code <= "111101";	-- 5
		elsif (color_value = "11110") then color_code <= "111111";	-- 
		else                               color_code <= "111111";	-- white
		end if;
		
	 
	end if;
	
end process;


mk_chcols  : process (PIXEL_CLOCK)
begin

	if rising_edge(PIXEL_CLOCK) then

		if (image_count(0) = '1') then
			VIDEO_R <= color_code(0) & color_code(0) & color_code(0) & color_code(0) & color_code(0);
			VIDEO_G <= color_code(1) & color_code(1) & color_code(1) & color_code(1) & color_code(1);
			VIDEO_B <= color_code(2) & color_code(2) & color_code(2) & color_code(2) & color_code(2);
			
		else
			VIDEO_R <= color_code(3) & color_code(3) & color_code(3) & color_code(3) & color_code(3);
			VIDEO_G <= color_code(4) & color_code(4) & color_code(4) & color_code(4) & color_code(4);
			VIDEO_B <= color_code(5) & color_code(5) & color_code(5) & color_code(5) & color_code(5);
			
		end if;
		
	end if;
	
end process;



  
------------------
--synch delay
------------------

p_synch_mux : process(PIXEL_CLOCK)
begin
	if rising_edge (PIXEL_CLOCK) then
	
		if (RESET = '1') then

			h_sync_d1 <= '1';
			v_sync_d1 <= '1';
			h_sync_d2 <= '1';
			v_sync_d2 <= '1';
			h_sync_d3 <= '1';
			v_sync_d3 <= '1';
			h_sync_d4 <= '1';
			v_sync_d4 <= '1';
			h_sync_d5 <= '1';
			v_sync_d5 <= '1';
			
			VSYNC <= '1';
			HSYNC <= '1';
			
		else
		
			-- perform delay to compensate calculation latency
			h_sync_d1 <= h_sync;
			v_sync_d1 <= v_sync;
			h_sync_d2 <= h_sync_d1;
			v_sync_d2 <= v_sync_d1;
			h_sync_d3 <= h_sync_d2;
			v_sync_d3 <= v_sync_d2;
			h_sync_d4 <= h_sync_d3;
			v_sync_d4 <= v_sync_d3;
			h_sync_d5 <= h_sync_d4;
			v_sync_d5 <= v_sync_d4;
			
			HSYNC <= h_sync_d5;
			VSYNC <= v_sync_d5;
			
		end if;		

	end if;

end process;
  
end architecture a;

